@import url("https://fonts.googleapis.com/css2?family=Inter:wght@400;600&display=swap");

html,
body {
  padding: 0;
  margin: 0;
}
:focus {
  outline: 0;
}
body {
  background: pink;
  color: #f5f3f5;
  font-family: "Inter", sans-serif;
  font-size: 16px;
}
form {
  background: pink;
  width: 600px;
  margin: 0 auto;
  padding: 2.5rem;
  box-sizing: border-box;
  font-size: 0.875rem;
}
form h3 {
  font-size: 1.5rem;
  margin-top: 0;
}
form p {
  margin-bottom: 0.625rem;
  line-height: 120%;
}
input[type="text"],
input[type="email"],
input[type="password"],
input[type="number"],
input[type="search"],
textarea,
select {
  width: 100%;
  max-width: 100%;
}
input[type="date"],
input[type="datetime-local"],
input[type="time"],
input[type="number"] {
  width: 200px;
}
input,
textarea,
select {
  background: #120912;
  border: 1px solid #4d414d;
  border-radius: 2px;
  -webkit-border-radius: 2px;
  padding: 0 12px;
  margin-bottom: 1rem;
  margin-top: 0.625rem;
  color: inherit;
  font-family: inherit;
}
input,
select {
  height: 2rem;
  min-height: 2rem;
}
input[type="radio"],
input[type="checkbox"] {
  height: auto;
  min-height: auto;
}
textarea {
  height: 6rem;
  min-height: 6rem;
  padding: 12px;
  vertical-align: top;
  overflow: auto;
}
input:focus,
textarea:focus {
  box-shadow: rgb(18, 9, 18) 0px 0px 0px 2px, rgb(120, 80, 250) 0px 0px 0px 4px;
}
form select {
  -webkit-appearance: button;
  border-radius: 2px;
  -webkit-border-radius: 2px;
  -webkit-user-select: none;
  background-image: url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' width='18' height='18' viewBox='0 0 24 24' fill='currentColor' stroke='none'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M5.293 8.293a1 1 0 011.414 0L12 13.586l5.293-5.293a1 1 0 111.414 1.414l-6 6a1 1 0 01-1.414 0l-6-6a1 1 0 010-1.414z' fill='currentColor'%3E%3C/path%3E%3C/svg%3E");
  background-position: center right;
  background-repeat: no-repeat;
  font-size: inherit;
  overflow: hidden;
  text-overflow: ellipsis;
  white-space: nowrap;
}
input[type="submit"] {
  -webkit-box-align: center;
  align-items: center;
  border-radius: 2px;
  -webkit-border-radius: 2px;
  padding: 0px 16px;
  height: 2rem;
  line-height: 14px;
  background: rgb(120, 80, 250);
  margin-bottom: 0;
  border: none;
  margin-right: 20px;
}
input[type="submit"]:hover {
  background: linear-gradient(rgb(120, 80, 250) 0%, rgb(98, 65, 204) 100%);
  cursor: pointer;
}
